from threading import Lock, Event
from datetime import timedelta
from typing import Union, List, Any, Callable
from inspect import getframeinfo, stack
from time import sleep
from torch.distributed import rpc
from machin.parallel.exception import ExceptionWithTraceback
from machin.utils.logging import default_logger
from logging import DEBUG

import enum
import torch as t
import torch.distributed as dist
import torch.distributed.distributed_c10d as dist_c10d

WORLD = None  # type: Union[None, World]


class LUTType(enum.Enum):
    VALUE = 0
    SERVICE = 1


def debug_with_process(message):
    if default_logger.level != DEBUG:
        return
    caller = getframeinfo(stack()[1][0])
    default_logger.debug(
        f"Process [{get_cur_rank()}]: "
        f"<{caller.filename},L{caller.lineno}> "
        f"{message}"
    )


def _copy_doc(from_func):
    """
    Used by collective group to copy documents from torch.
    """
    import io
    import sys

    def _decorator(func):
        if "sphinx" in sys.modules:  # pragma: no cover
            src_doc = from_func.__doc__
            lines = io.StringIO(src_doc)
            # remove the group line
            src_doc = "".join([line for line in lines if "group" not in line])
            func.__doc__ = src_doc
        return func

    return _decorator


def _rpc_set_lut_entry(group_name, key, proc_name, lut_type):  # pragma: no cover
    table = WORLD.value_lut if lut_type == LUTType.VALUE else WORLD.service_lut
    with WORLD.lut_lock:
        if (group_name, key) in table:
            return False
        else:
            table[(group_name, key)] = proc_name
            return True


def _rpc_unset_lut_entry(group_name, key, proc_name, lut_type):  # pragma: no cover
    table = WORLD.value_lut if lut_type == LUTType.VALUE else WORLD.service_lut
    with WORLD.lut_lock:
        if (group_name, key) in table:
            if table[(group_name, key)] == proc_name:
                table.pop((group_name, key))
                return True
        return False


def _rpc_get_lut_entry(group_name, key, lut_type):  # pragma: no cover
    table = WORLD.value_lut if lut_type == LUTType.VALUE else WORLD.service_lut
    with WORLD.lut_lock:
        if (group_name, key) in table:
            return True, table[(group_name, key)]
        else:
            return False, None


def _rpc_has_lut_entry(group_name, key, lut_type):  # pragma: no cover
    table = WORLD.value_lut if lut_type == LUTType.VALUE else WORLD.service_lut
    with WORLD.lut_lock:
        if (group_name, key) in table:
            return True
        else:
            return False


def _rpc_call_func(func, args, kwargs):  # pragma: no cover
    # Call a function/bound method
    try:
        return func(*args, **kwargs)
    except BaseException as e:
        raise RpcException(e)


def _rpc_call_service(group_name, key, args, kwargs):  # pragma: no cover
    # call a registered service
    world = get_world()
    if group_name not in world.groups:
        # could happen if group has been destroyed on this process
        # deregister the entry from lut manager
        rpc.rpc_sync(
            world.lut_manager,
            _rpc_unset_lut_entry,
            args=(group_name, key, get_cur_name(), LUTType.SERVICE),
        )
        raise KeyError(f"Group [{group_name}], not found on Process [{get_cur_name()}]")
    lut = WORLD.groups[group_name].group_service_lut

    if key in lut:
        try:
            return lut[key](*args, **kwargs)
        except BaseException as e:
            raise RpcException(e)
    else:
        # could happen if local map is not synchronized with the
        # global map
        # deregister the entry from lut manager
        rpc.rpc_sync(
            world.lut_manager,
            _rpc_unset_lut_entry,
            args=(group_name, key, get_cur_name(), LUTType.VALUE),
        )
        raise KeyError(
            f"Service [{key}] not found on Group [{group_name}], "
            f"Process [{get_cur_name()}]"
        )


def _rpc_get_paired_value(group_name, key):  # pragma: no cover
    # get a paired value
    world = get_world()
    if group_name not in world.groups:
        # could happen if group has been destroyed on this process
        # deregister the entry from lut manager
        rpc.rpc_sync(
            world.lut_manager,
            _rpc_unset_lut_entry,
            args=(group_name, key, get_cur_name(), LUTType.VALUE),
        )
        raise KeyError(f"Group [{group_name}], not found on Process [{get_cur_name()}]")

    paired_map = WORLD.groups[group_name].group_value_lut

    if key in paired_map:
        return paired_map[key]
    else:
        # could happen if local map is not synchronized with the
        # global map
        # deregister the entry from lut manager
        rpc.rpc_sync(
            world.lut_manager,
            _rpc_unset_lut_entry,
            args=(group_name, key, get_cur_name(), LUTType.VALUE),
        )
        raise KeyError(
            f"Value with key [{key}] not found on Group [{group_name}], "
            f"Process [{get_cur_name()}]"
        )


def _world_singleton(cls):
    # Decorator used to make sure that there is only one world instance.
    def _world_singleton_wrapper(*args, **kwargs):
        global WORLD
        if WORLD is None:
            WORLD = cls(*args, **kwargs)
        else:  # pragma: no cover
            raise RuntimeError("World could only be initialized once!")
        return WORLD

    return _world_singleton_wrapper


def _torch_version_less_than(major, minor):
    t_ver = [int(v) for v in t.__version__.split(".")[0:2]]
    return t_ver < [major, minor]


def get_cur_rank():
    """
    Returns:
        Current real process rank.
    """
    if WORLD is None:  # pragma: no cover
        raise RuntimeError("Distributed environment not initialized!")
    return WORLD.rank


def get_cur_name():
    """
    Returns:
        Current real process name.
    """
    if WORLD is None:  # pragma: no cover
        raise RuntimeError("Distributed environment not initialized!")
    return WORLD.name


def get_world():  # pragma: no cover
    return WORLD


def is_world_initialized():  # pragma: no cover
    return WORLD is not None


def _is_group_ready(group_name):  # pragma: no cover
    return WORLD.group_create_signals.get(group_name, None) is False


def _unlock_group(group_name):  # pragma: no cover
    WORLD.group_create_signals[group_name] = True


def _check_executor(func):
    def wrapped(self, *args, **kwargs):
        if get_cur_name() not in self.group_members:
            raise RuntimeError(
                f"You should not execute function {func.__qualname__} when "
                "current process is not a member of the group"
            )
        return func(self, *args, **kwargs)

    return wrapped


class RpcException(Exception):  # pragma: no cover
    """
    Rpc exception class.
    """

    def __init__(self, msg):
        if isinstance(msg, str):
            # used by rpc when reraising the exception on the caller side
            super().__init__(msg)
        else:
            tb = ExceptionWithTraceback(msg).tb
            super().__init__(tb)


@_world_singleton
class World:
    """
    The distributed world.
    """

    def __init__(
        self,
        name: str,
        rank: int = -1,
        world_size: int = -1,
        init_dist: bool = True,
        init_rpc: bool = True,
        dist_backend: str = "gloo",
        dist_init_method: str = "tcp://localhost:9100",
        rpc_init_method: str = "tcp://localhost:9101",
        dist_timeout: float = 60,
        rpc_timeout: float = 60,
    ):
        """
        Args:
            name: A unique name to identify current process.
            rank: A unique rank of the current process. You do not need to specify
                it if you are using `torch.distributed.launch` or `torchelastic`
            world_size:   Size of the distributed world. You do not need to specify
                it if you are using `torch.distributed.launch` or `torchelastic`
            dist_timeout: Distributed package timeout in seconds.
            rpc_timeout:  Global rpc call timeout in seconds.
        """
        self.world_size = world_size
        self.rank = rank
        self.name = name
        self.groups = {}
        self.group_create_signals = {}

        if init_dist:
            dist.init_process_group(
                backend=dist_backend,
                init_method=dist_init_method,
                timeout=timedelta(seconds=dist_timeout),
                rank=rank,
                world_size=world_size,
            )
        if init_rpc:
            rpc.init_rpc(
                self.name,
                rank=rank,
                world_size=world_size,
                rpc_backend_options=rpc.TensorPipeRpcBackendOptions(
                    init_method=rpc_init_method, rpc_timeout=rpc_timeout
                ),
            )

        # get rank-name mapping
        self.rank_name_map = {}
        for wi in rpc._get_current_rpc_agent().get_worker_infos():
            self.rank_name_map[wi.id] = wi.name

        # Start role dispatching.
        self.started = True
        self.rpc_timeout = rpc_timeout

        # map for paired values and registered services
        self.value_lut = {}
        self.service_lut = {}
        self.lut_lock = Lock()
        self.lut_manager = self.rank_name_map[0]

    def stop(self):  # pragma: no cover
        if not self.started:
            raise RuntimeError("Cannot stop the world multiple times!")
        else:
            rpc.shutdown()

    def create_collective_group(
        self, ranks: List[int], timeout: float = 60, backend: Any = None,
    ):
        """
        Create a sub process group for collective communications. This function
        is blocking and requires that all processes in ``ranks`` to
        enter this function.

        Warning:
            Do not make collective communications call in sub-processes,
            it is unsafe.

        Args:
            ranks: Ranks of involved processes.
            timeout: Timeout of operations in the new group.
            backend: New group backend.
        Returns:
            A ``Group`` with type ``Group.COLLECTIVE``
        """
        ranks = sorted(ranks)
        group = CollectiveGroup(
            dist.new_group(ranks, timedelta(seconds=timeout), backend),
            ranks.index(self.rank) if self.rank in ranks else None,
        )
        return group

    def create_rpc_group(self, group_name: str, members: List[str]):
        """
        Create a sub process group for rpc calls. This function
        is blocking and requires that all processes in ``members`` to
        enter this function.

        Args:
            group_name: A unique group name.
            members: Members of the group.

        Returns:
            A rpc group.
        """
        if get_cur_name() not in members:  # pragma: no cover
            raise RuntimeError(
                f"Creator Process [{get_cur_name()}] not in Group [{group_name}]"
            )
        if group_name in self.groups:  # pragma: no cover
            raise RuntimeError(f"Group {group_name} already exists!")
        group = RpcGroup(group_name, members)

        # set the group
        self.groups[group_name] = group

        # temporarily set a signal
        self.group_create_signals[group_name] = False
        # wait for other members to enter
        if get_cur_name() == members[0]:
            while True:
                sleep(0.1)
                future = [
                    rpc.rpc_async(m, _is_group_ready, args=(group_name,))
                    for m in members
                ]
                for fut in future:
                    if not fut.wait():
                        break
                else:
                    future = [
                        rpc.rpc_async(m, _unlock_group, args=(group_name,))
                        for m in members
                    ]
                    for fut in future:
                        fut.wait()
                    # finish syncing all processes
                    break
        else:
            while self.group_create_signals[group_name] is not True:
                sleep(0.1)
        return group

    def get_ranks(self):
        """
        Returns:
            A list of ranks of all processes.
        """
        return list(self.rank_name_map.keys())

    def get_members(self):
        """
        Returns:
            A list of names of all processes.
        """
        return list(self.rank_name_map.values())

    def __reduce__(self):  # pragma: no cover
        raise RuntimeError("World is not picklable, create it per process!")


class CollectiveGroup:
    """
    A thin wrapper of collective communication primitives of
    ``torch.distributed``, the only difference is that ``irecv``
    now supports to recv from any src
    """

    def __init__(self, group, current_relative_rank):
        """
        Do not do it your self, use :meth:`~machin.parallel\
.distributed.World.create_collective_group` instead.
        """
        self.group = group
        self.current_rank = current_relative_rank
        self.destroyed = False

    @_copy_doc(dist.send)
    def send(self, tensor, dst, tag=0):
        return dist.send(tensor, dst, self.group, tag)

    @_copy_doc(dist.recv)
    def recv(self, tensor, src=None, tag=0):
        return dist.recv(tensor, src, self.group, tag)

    @_copy_doc(dist.isend)
    def isend(self, tensor, dst, tag=0):
        return dist.isend(tensor, dst, self.group, tag)

    def irecv(self, tensor, src=None, tag=0):  # pragma: no cover
        """
        Returns:
            An object you can call .wait() on, .wait()
            will return the source rank.
        """
        # pylint: disable=protected-access

        # Original irecv doesn't support recv from any
        # but original recv does. They are essentially
        # the same except recv have a wait() call
        dist_c10d._check_single_tensor(tensor, "tensor")
        if dist_c10d._rank_not_in_group(self.group):

            class Waiter:
                def wait(self):
                    return -1

            return Waiter()

        if self.group == dist_c10d.GroupMember.WORLD:
            dist_c10d._check_default_pg()
            pg = dist_c10d._default_pg
        else:
            pg = self.group

        if src is None:
            work = pg.recv_anysource([tensor], tag)
            if self.group == dist_c10d.GroupMember.WORLD:

                class Waiter:
                    def wait(self):
                        nonlocal work
                        work.wait()
                        return work.source_rank()

                return Waiter()
            else:

                class Waiter:
                    def wait(self):
                        nonlocal work, pg
                        work.wait()
                        src_rank = work.source_rank()
                        return dist_c10d._get_global_rank(pg, src_rank)

                return Waiter()
        else:
            if self.group == dist_c10d.GroupMember.WORLD:
                work = pg.recv([tensor], src, tag)
            else:
                group_src_rank = dist_c10d._get_group_rank(pg, src)
                work = pg.recv([tensor], group_src_rank, tag)

            class Waiter:
                def wait(self):
                    nonlocal src
                    work.wait()
                    return src

            return Waiter()

    @_copy_doc(dist.broadcast)
    def broadcast(self, tensor, src, async_op=False):
        return dist.broadcast(tensor, src, self.group, async_op)

    @_copy_doc(dist.all_reduce)
    def all_reduce(self, tensor, op=dist.ReduceOp.SUM, async_op=False):
        return dist.all_reduce(tensor, op, self.group, async_op)

    @_copy_doc(dist.reduce)
    def reduce(self, tensor, dst, op=dist.ReduceOp.SUM, async_op=False):
        return dist.reduce(tensor, dst, op, self.group, async_op)

    @_copy_doc(dist.all_gather)
    def all_gather(self, tensor_list, tensor, async_op=False):
        return dist.all_gather(tensor_list, tensor, self.group, async_op)

    @_copy_doc(dist.gather)
    def gather(self, tensor, gather_list, dst=0, async_op=False):
        return dist.gather(tensor, gather_list, dst, self.group, async_op)

    @_copy_doc(dist.scatter)
    def scatter(self, tensor, scatter_list=None, src=0, async_op=False):
        return dist.scatter(tensor, scatter_list, src, self.group, async_op)

    @_copy_doc(dist.barrier)
    def barrier(self, async_op=False):
        return dist.barrier(self.group, async_op)

    @_copy_doc(dist.broadcast_multigpu)
    def broadcast_multigpu(self, tensor_list, src, async_op=False, src_tensor=0):
        return dist.broadcast_multigpu(
            tensor_list, src, self.group, async_op, src_tensor
        )

    @_copy_doc(dist.all_reduce_multigpu)
    def all_reduce_multigpu(self, tensor_list, op=dist.ReduceOp.SUM, async_op=False):
        return dist.all_reduce_multigpu(tensor_list, op, self.group, async_op)

    @_copy_doc(dist.reduce_multigpu)
    def reduce_multigpu(
        self, tensor_list, dst, op=dist.ReduceOp.SUM, async_op=False, dst_tensor=0
    ):  # pragma: no cover
        return dist.reduce_multigpu(
            tensor_list, dst, op, self.group, async_op, dst_tensor
        )

    @_copy_doc(dist.all_gather_multigpu)
    def all_gather_multigpu(
        self, output_tensor_lists, input_tensor_list, async_op=False
    ):  # pragma: no cover
        return dist.all_gather_multigpu(
            output_tensor_lists, input_tensor_list, self.group, async_op
        )

    def destroy(self):
        """
        Destroy this collective communication group.
        """
        if not self.destroyed:
            dist.destroy_process_group(self.group)
            self.destroyed = True

    def size(self):
        """
        Returns:
            collective group size.
        """
        return dist.get_world_size(self.group)

    def __reduce__(self):  # pragma: no cover
        raise RuntimeError("Group is not picklable, create it per process!")

    def __del__(self):
        self.destroy()


# TODO:
# add the heartbeat mechanism to the lut_manager, to increase robustness


class RpcGroup:
    def __init__(self, group_name, group_members, first_create=True):
        self.group_name = group_name
        self.group_members = group_members
        self.group_value_lut = {}
        self.group_service_lut = {}
        self.destroyed = False
        self._barrier_event = Event()
        self._barrier_status = False
        if first_create and self.is_member(get_cur_name()):
            self.register(
                f"_rpc_entered_barrier_{get_cur_name()}", self._rpc_entered_barrier,
            )
            self.register(f"_rpc_exit_barrier_{get_cur_name()}", self._rpc_exit_barrier)

    @_copy_doc(rpc.rpc_sync)
    def rpc_sync(self, to: str, func: Callable, timeout=-1, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        return self._rpc_normal_call(rpc.rpc_sync, to, func, timeout, args, kwargs)

    @_copy_doc(rpc.rpc_async)
    def rpc_async(self, to: str, func: Callable, timeout=-1, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        return self._rpc_normal_call(rpc.rpc_async, to, func, timeout, args, kwargs)

    @_copy_doc(rpc.remote)
    def remote(self, to: str, func: Callable, timeout=-1, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        return self._rpc_normal_call(rpc.remote, to, func, timeout, args, kwargs)

    @_check_executor
    def pair(self, key: Any, value: Any):
        """
        Pair a value to current process group.

        Args:
            key: A key which uniquely identifies this value in this group.
                 The name only needs to be unique for this value in this
                 group.

            value: Value to be paired.

        Raise:
            ``KeyError`` if value has already been paired.
        """
        if key in self.group_value_lut:
            raise KeyError(
                f'Value with key "{key}" already paired to Group [{self.group_name}]'
            )
        # announce the pairing
        status = rpc.rpc_sync(
            get_world().lut_manager,
            _rpc_set_lut_entry,
            args=(self.group_name, key, get_cur_name(), LUTType.VALUE),
        )
        if status:
            self.group_value_lut[key] = value
        else:
            raise KeyError(
                f'Value with key "{key}" already paired to Group [{self.group_name}]'
            )

    @_check_executor
    def unpair(self, key: Any):
        """
        Unpair a paired value from current process group. The key must be
        paired by the current process.

        Args:
            key: A key which uniquely identifies this value in this group.
                 The name only needs to be unique for this value in this
                 group.
        Raise:
            ``KeyError`` if value has not been paired.
        """
        if key not in self.group_value_lut:
            raise KeyError(
                f'Value with key "{key}" not paired to Group [{self.group_name}] '
                f"on Process[{get_cur_name()}]"
            )
        # announce the unpairing
        status = rpc.rpc_sync(
            get_world().lut_manager,
            _rpc_unset_lut_entry,
            args=(self.group_name, key, get_cur_name(), LUTType.VALUE),
        )
        if status:
            self.group_value_lut.pop(key)
        else:  # pragma: no cover
            # should never happen
            raise RuntimeError(
                f'Failed to unpair value with key "{key}" from '
                f"Group [{self.group_name}], executor is Process[{get_cur_name()}]"
            )

    def is_paired(self, key: Any):
        """
        Check whether a key has been paired to the current group.

        Args:
            key: A key which uniquely identifies this value in this group.
                 The name only needs to be unique for this value in this
                 group.
        """
        return rpc.rpc_sync(
            get_world().lut_manager,
            _rpc_has_lut_entry,
            args=(self.group_name, key, LUTType.VALUE),
        )

    def get_paired(self, key: Any):
        """
        Args:
            key: Key of the paired value, in this group.

        Returns:
            A RRef to the paired value.

        Raises:
            ``KeyError`` if not found.
        """
        if key in self.group_value_lut:
            holder = get_cur_name()
        else:
            status, holder = rpc.rpc_sync(
                get_world().lut_manager,
                _rpc_get_lut_entry,
                args=(self.group_name, key, LUTType.VALUE),
            )
            if not status:
                raise KeyError(
                    f"Value with key [{key}] not found on Group [{self.group_name}], "
                )
        return rpc.remote(holder, _rpc_get_paired_value, args=(self.group_name, key))

    @_check_executor
    def register(self, key: Any, service: Any):
        """
        Register a service to current process group.

        Args:
            key: A key which uniquely identifies this service in this group.
                 The name only needs to be unique for this service in this
                 group.
            service: Service to be registered.

        Raise:
            ``KeyError`` if service has already been registered.
        """
        if key in self.group_service_lut:
            raise KeyError(
                f'Service with key "{key}" already registered '
                f"in Group [{self.group_name}]"
            )
        # announce the pairing
        status = rpc.rpc_sync(
            get_world().lut_manager,
            _rpc_set_lut_entry,
            args=(self.group_name, key, get_cur_name(), LUTType.SERVICE),
        )
        if status:
            self.group_service_lut[key] = service
        else:
            raise KeyError(
                f'Service with key "{key}" already registered '
                f"in Group [{self.group_name}]"
            )

    @_check_executor
    def deregister(self, key: Any):
        """
        Deregister service from current process group. The key must be
        paired by the current process.

        Args:
            key: A key which uniquely identifies this value in this group.
                 The name only needs to be unique for this value in this
                 group.
        Raise:
            ``KeyError`` if srvice has not been registered.
        """
        if key not in self.group_service_lut:
            raise KeyError(
                f'Service with key "{key}" not registered '
                f"in Group [{self.group_name}] "
                f"on Process[{get_cur_name()}]"
            )
        # announce the deregistration
        status = rpc.rpc_sync(
            get_world().lut_manager,
            _rpc_unset_lut_entry,
            args=(self.group_name, key, get_cur_name(), LUTType.SERVICE),
        )
        if status:
            self.group_service_lut.pop(key)
        else:  # pragma: no cover
            # should never happen
            raise RuntimeError(
                f'Failed to deregister service with key "{key}" '
                f"from Group [{self.group_name}], "
                f"executor is Process[{get_cur_name()}]"
            )

    def is_registered(self, key: Any):
        """
        Check whether a service has been registered in the current group.

        Args:
            key: A key which uniquely identifies this service in this group.
                 The name only needs to be unique for this service in this
                 group.
        """
        return rpc.rpc_sync(
            get_world().lut_manager,
            _rpc_has_lut_entry,
            args=(self.group_name, key, LUTType.SERVICE),
        )

    def registered_sync(self, key: Any, args=(), kwargs=None):
        """
        Args:
            key: Key of the registered service, in this group.
            args: Service arguments.
            kwargs: Service keyword arguments.

        Returns:
            Result returned by the service.

        Raises:
            ``KeyError`` if service is not found.
        """
        if kwargs is None:
            kwargs = {}
        return self._rpc_service_call(rpc.rpc_sync, key, args, kwargs)

    def registered_async(self, key: Any, args=(), kwargs=None):
        """
        Args:
            key: Key of the registered service, in this group.
            args: Service arguments.
            kwargs: Service keyword arguments.

        Returns:
            A future object you can call ``wait()``on.
            ``wait()`` will block the thread until execution is completed,
            and will return the result returned by the service.

        Raises:
            ``KeyError`` if service is not found.
        """
        if kwargs is None:
            kwargs = {}
        return self._rpc_service_call(rpc.rpc_async, key, args, kwargs)

    def registered_remote(self, key: Any, args=(), kwargs=None):
        """
        Args:
            key: Key of the registered service, in this group.
            args: Service arguments.
            kwargs: Service keyword arguments.

        Returns:
            A RRef object pointing to the result returned by the service.

        Raises:
            ``KeyError`` if service is not found.
        """
        if kwargs is None:
            kwargs = {}
        return self._rpc_service_call(rpc.remote, key, args, kwargs)

    @_check_executor
    def barrier(self):
        """
        Synchronize all members in the group, until all members have entered
        a ``barrier()`` function.

        Not thread-safe.
        """

        self._barrier_status = True
        if get_cur_name() == self.group_members[0]:
            while True:
                all_entered = all(
                    self.registered_sync(f"_rpc_entered_barrier_{m}")
                    for m in self.group_members
                )
                if not all_entered:
                    sleep(0.2)
                else:
                    break
            for m in self.group_members:
                self.registered_sync(f"_rpc_exit_barrier_{m}")
        else:
            self._barrier_event.wait()

    @_check_executor
    def destroy(self):
        """
        Destroy the rpc group.

        Note: deregistration is not considered, because they will be purged
            when any lookup fail.
        """
        if not self.destroyed:
            WORLD.groups.pop(self.group_name)
            self.destroyed = True

    def size(self):
        """
        Get the number of members in group.
        """
        return len(self.group_members)

    def is_member(self, target: str = None) -> bool:
        """
        Check whether target name is a group member.
        """
        if target is None:
            target = self.get_cur_name()
        return target in self.group_members

    def get_group_name(self) -> str:
        """
        Returns:
            Name of this group.
        """
        return self.group_name

    def get_group_members(self) -> List[str]:
        """
        Returns:
            A list of group members.
        """
        return self.group_members

    @staticmethod
    def get_cur_name() -> str:
        return get_world().name

    def _rpc_normal_call(self, rpc_method, to, func, timeout, args, kwargs):
        if not self.is_member(to):  # pragma: no cover
            raise RuntimeError("RPC target is not a member of group.")

        new_args = (func, args, kwargs)
        if _torch_version_less_than(1, 6):
            return rpc_method(to, _rpc_call_func, args=new_args)
        return rpc_method(to, _rpc_call_func, args=new_args, timeout=timeout)

    def _rpc_service_call(self, rpc_method, key, args, kwargs):
        if key in self.group_service_lut:
            holder = get_cur_name()
        else:
            status, holder = rpc.rpc_sync(
                get_world().lut_manager,
                _rpc_get_lut_entry,
                args=(self.group_name, key, LUTType.SERVICE),
            )
            if not status:
                raise KeyError(
                    f"Service with key [{key}] not found on Group [{self.group_name}], "
                )
        return rpc_method(
            holder, _rpc_call_service, args=(self.group_name, key, args, kwargs)
        )

    def _rpc_entered_barrier(self):
        return self._barrier_status

    def _rpc_exit_barrier(self):
        self._barrier_status = False
        self._barrier_event.set()
        self._barrier_event.clear()

    def __reduce__(self):  # pragma: no cover
        # returns a complete description of group
        return RpcGroup, (self.group_name, self.group_members, False)



# Adapted from https://github.com/modestyachts/ARS

from typing import Union, Dict, List, Tuple, Callable, Any
import copy
import torch as t
import torch.nn as nn
import numpy as np

from machin.model.nets.base import NeuralNetworkModule
from machin.parallel.server import PushPullModelServer
from machin.parallel.distributed import RpcGroup, get_world
from machin.frame.helpers.servers import model_server_helper
from machin.utils.logging import default_logger
from .base import TorchFramework, Config
from .utils import (
    safe_call,
    safe_return,
    assert_and_get_valid_models,
    assert_and_get_valid_optimizer,
    assert_and_get_valid_lr_scheduler,
)


class RunningStat:
    """
    Running status estimator method by B. P. Welford
    described in http://www.johndcook.com/blog/standard_deviation/
    """

    def __init__(self, shape):
        """
        Create a running status (mean, var, std) estimator.

        Note:
            Running on CPU. Input tensors are also required to locate
            on CPU.

        Args:
            shape: Shape of input elements.
        """
        # number of pushed samples
        self._n = 0
        # mean
        self._M = t.zeros(shape, dtype=t.float64)
        # variance * (_n - 1)
        self._S = t.zeros(shape, dtype=t.float64)

    def copy(self):
        """
        Returns a copy of the running status estimator.
        """
        other = RunningStat(self.shape)
        other._n = self._n
        other._M = self._M.clone()
        other._S = self._S.clone()
        return other

    def push(self, x: t.Tensor):
        """
        Add a new sample to the running status estimator.

        Args:
            x: New sample.
        """
        assert x.dtype == t.float64, "Input are required to be float64!"
        assert x.shape == self._M.shape, "Shape mismatch!"
        assert x.device == self._M.device, "Device mismatch!"
        n_old = self._n
        self._n += 1
        if self._n == 1:
            self._M.copy_(x)
        else:
            delta = x - self._M
            self._M += delta / self._n
            self._S += delta * delta * n_old / self._n

    def update(self, other: "RunningStat"):
        """
        combine this estimator with another estimator.

        Args:
            other: Another running status estimator, with the same shape.
        """
        # combined variance:
        # https://www.emathzone.com/tutorials/basic-statistics/
        # combined-variance.html

        # Note: S is variance multiplied by n - 1. you need to
        # deduce a little bit to get the calculation method of S
        assert other.shape == self.shape, "Shape mismatch!"
        n1 = self._n
        n2 = other._n
        n = n1 + n2
        delta = self._M - other._M
        delta2 = delta * delta
        M = (n1 * self._M + n2 * other._M) / n
        S = self._S + other._S + delta2 * n1 * n2 / n
        self._n = n
        self._M = M
        self._S = S

    def __repr__(self):
        return (
            f"RunningStat(shape={self._M.shape}, n={self.n}, "
            f"mean_mean={t.mean(self.mean)}, mean_std={t.mean(self.std)})"
        )

    @property
    def n(self):
        return self._n

    @property
    def mean(self):
        return self._M

    @property
    def var(self):
        # at self._N = 0
        # return self._M instead of self._S
        # to prevent division by 0 in state normalization
        return self._S / (self._n - 1) if self._n > 1 else t.square(self._M)

    @property
    def std(self):
        return t.sqrt(self.var)

    @property
    def shape(self):
        return self._M.shape


class MeanStdFilter:
    """Keeps track of a running mean for seen states"""

    def __init__(self, shape):
        """
        Create a state normalization filter.

        Note:
            Running on CPU. Input tensors are also required to locate
            on CPU.

        Args:
            shape: Shape of the state
        """
        self.shape = shape
        self.rs = RunningStat(shape)

        # In distributed rollouts, each worker sees different states.
        # `rs_local` is used to keep track of encountered local states.
        # In each rollout episode, the manager process will collect `rs_local`
        # from all workers, then distribute new running states to `rs` of all
        # workers.

        self.rs_local = RunningStat(shape)

        self.mean = t.zeros(shape, dtype=t.float64)
        self.std = t.ones(shape, dtype=t.float64)

    def clear_local(self):
        self.rs_local = RunningStat(self.shape)
        return

    def copy(self):
        """
        Returns a copy of Filter.
        """
        other = MeanStdFilter(self.shape)
        other.rs = self.rs.copy()
        other.rs_local = self.rs_local.copy()
        return other

    def collect(self, other: "MeanStdFilter"):
        """
        Takes a worker's filter and collect their local status.

        Using notation `F(state, buffer)`, given `Filter1(x1, y1)` and
        `Filter2(x2, yt)`, `Filter1` to `Filter1(x1 + yt, y1)`.

        This function is **only** used by the manager process.
        """
        self.rs.update(other.rs_local)
        return

    def sync(self, other):
        """
        Syncs all fields together from other filter.

        Using notation `F(state, buffer)` Given `Filter1(x1, y1)` and
        `Filter2(x2, yt)`, `sync` modifies `Filter1` to `Filter1(x2, yt)`

        This function is used by all worker processes.
        """
        assert other.shape == self.shape, "Shape mismatch!"
        self.rs = other.rs.copy()
        self.rs_local = other.rs_local.copy()
        return

    def apply_stats(self):
        """
        Apply the mean and std stored in `self.rs` to normalization.
        """
        self.mean = self.rs.mean
        self.std = self.rs.std

        # Set values for std less than 1e-7 to +inf to avoid
        # dividing by zero. State elements with zero variance
        # are set to zero as a result.
        self.std[self.std < 1e-7] = float("inf")
        return

    def filter(self, x: t.Tensor, update: bool = True):
        """
        Filter(normalize) observed state.

        Note:
            `update` will only update internal records of mean and std
            of the filter stored in `self.rs` and `self.rs_local`,
            but will not update the used mean and std in normalization.

            You must call `apply_stats()` to apply mean and std from
            `self.rs`.

        Args:
            x: State to filter.
            update: Whether update the filter with the current state.

        Returns:
            Normalized state.
        """
        if update:
            self.rs.push(x)
            self.rs_local.push(x)
        x = x - self.mean
        x = x / (self.std + 1e-8)
        return x

    def __repr__(self):
        return (
            f"MeanStdFilter(shape={self.shape}, rs={self.rs}, "
            f"rs_local={self.rs_local})"
        )


class SharedNoiseSampler:
    def __init__(self, noise: t.Tensor, seed: int):
        """
        Args:
            noise: Noise tensor to sample from.
            seed: Index random sampling seed.
        """
        self.rg = np.random.RandomState(seed)
        self.noise = noise
        assert self.noise.dtype == t.float64

    def get(self, idx, size):
        return self.noise[idx : idx + size]

    def sample(self, size):
        """
        Args:
            size: Size of sampled 1D noise tensor.

        Returns:
            Noise begin index, noise tensor.
        """
        idx = self.rg.randint(0, len(self.noise) - size + 1)
        return idx, self.noise[idx : idx + size]


class ARS(TorchFramework):
    """
    ARS framework.
    """

    _is_top = ["actor"]
    _is_restorable = ["actor"]

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        ars_group: RpcGroup,
        model_server: Tuple[PushPullModelServer],
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict] = None,
        learning_rate: float = 0.01,
        gradient_max: float = np.inf,
        noise_std_dev: float = 0.02,
        noise_size: int = 250000000,
        rollout_num: int = 32,
        used_rollout_num: int = 32,
        normalize_state: bool = True,
        noise_seed: int = 12345,
        sample_seed: int = 123,
        **__,
    ):
        """

        Note:
            The first process in `ars_group` will be the manager process.

        Args:
            actor: Actor network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            ars_group: Group of all processes using the ARS framework.
            model_server: Custom model sync server accessor for ``actor``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            gradient_max: Maximum gradient.
            noise_std_dev: Standard deviation of the shared noise array.
            noise_size: Size of the shared noise array.
            rollout_num: Number of rollouts executed by workers in group.
            used_rollout_num: Number of used rollouts.
            normalize_state:  Whether to normalize the state seen by actor.
            noise_seed: Random seed used to generate noise.
            sample_seed: Based random seed used to sample noise.
        """
        assert rollout_num >= used_rollout_num
        self.grad_max = gradient_max
        self.rollout_num = rollout_num
        self.used_rollout_num = used_rollout_num
        self.normalize_state = normalize_state
        self.ars_group = ars_group

        # determine the number of rollouts(pair of actors with neg/pos delta)
        # assigned to current worker process
        w_num = len(ars_group.get_group_members())
        w_index = ars_group.get_group_members().index(ars_group.get_cur_name())
        segment_length = int(np.ceil(rollout_num / w_num))
        self.local_rollout_min = w_index * segment_length
        self.local_rollout_num = min(
            segment_length, rollout_num - self.local_rollout_min
        )

        self.actor = actor
        # `actor_with_delta` use rollout index and delta sign as key.
        # where rollout index is the absolute global index of rollout
        # and delta sign is true for positive, false for negative
        self.actor_with_delta = {}  # type: Dict[Tuple[int, bool], t.nn.Module]
        self.actor_optim = optimizer(self.actor.parameters(), lr=learning_rate)
        self.actor_model_server = model_server[0]

        # `filter` use state name as key
        # eg: "state_1"
        self.filter = {}  # type: Dict[str, MeanStdFilter]

        # `delta_idx` use rollout index as key
        # The inner dict use model parameter name as key, and starting
        # noise index in the noise array as value.
        self.delta_idx = {}  # type: Dict[int, Dict[str, int]]

        # `reward` use rollout index as key, the first list stores
        # rewards of model with negative noise delta, the second list
        # stores rewards of model with positive noise delta.
        self.reward = {}  # type: Dict[int, Tuple[List, List]]

        if lr_scheduler is not None:
            if lr_scheduler_args is None:
                lr_scheduler_args = ((),)
            if lr_scheduler_kwargs is None:
                lr_scheduler_kwargs = ({},)
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim, *lr_scheduler_args[0], **lr_scheduler_kwargs[0],
            )

        # generate shared noise
        # estimate model parameter num first
        param_max_num = 0
        for param in actor.parameters():
            param_max_num = max(np.prod(np.array(param.shape)), param_max_num)
        if param_max_num * 10 > noise_size:
            default_logger.warning(
                "Maximum parameter size of your model is "
                f"{param_max_num}, which is more than 1/10 of your noise"
                f"size {noise_size}, consider increasing noise_size."
            )
        elif param_max_num >= noise_size:
            raise ValueError(
                f"Noise size {noise_size} is too small compared to"
                f"maximum parameter size {param_max_num}!"
            )

        # create shared noise array
        self.noise_array = t.tensor(
            np.random.RandomState(noise_seed).randn(noise_size).astype(np.float64)
            * noise_std_dev
        )

        # create a sampler for each parameter in each rollout model
        # key is model parameter name
        self.noise_sampler = {}  # type: Dict[int, Dict[str, SharedNoiseSampler]]
        param_num = len(list(actor.parameters()))
        for lrn in range(self.local_rollout_num):
            r_idx = lrn + self.local_rollout_min
            sampler = {}
            for p_idx, (name, param) in enumerate(actor.named_parameters()):
                # each model and its inner parameters use a different
                # sampling stream of the same noise array.
                sampler[name] = SharedNoiseSampler(
                    self.noise_array, sample_seed + r_idx * param_num + p_idx
                )
            self.noise_sampler[r_idx] = sampler

        # synchronize base actor parameters
        self._sync_actor()
        self._generate_parameter()
        self._reset_reward_dict()
        super().__init__()

    @property
    def optimizers(self):
        return [self.actor_optim]

    @optimizers.setter
    def optimizers(self, optimizers):
        self.actor_optim = optimizers[0]

    @property
    def lr_schedulers(self):
        if hasattr(self, "actor_lr_sch"):
            return [self.actor_lr_sch]
        return []

    @classmethod
    def is_distributed(cls):
        return True

    def get_actor_types(self) -> List[str]:
        """
        Returns:
            A list of actor types needed to be evaluated by current worker
            process.
        """
        names = [
            "positive_" + str(k[0]) if k[1] else "negative_" + str(k[0])
            for k in self.actor_with_delta.keys()
        ]
        return names

    def act(self, state: Dict[str, Any], actor_type: str, *_, **__):
        """
        Use actor network to give a policy to the current state.

        Args:
            state: State dict seen by actor.
            actor_type: Type of the used actor.

        Returns:
            Anything produced by actor.
        """
        # normalize states
        # filter shapes will be initialized on first call
        if self.normalize_state:
            for k, v in state.items():
                if k not in self.filter:
                    self.filter[k] = MeanStdFilter(v.shape)
                state[k] = (
                    self.filter[k]
                    .filter(v.to(dtype=t.float64, device="cpu"))
                    .to(dtype=v.dtype, device=v.device)
                )
        if actor_type == "original":
            return safe_return(safe_call(self.actor, state))
        elif actor_type.startswith("positive_") or actor_type.startswith("negative_"):
            rollout_idx = int(actor_type.split("_")[1])
            is_positive = actor_type[0] == "p"
            return safe_return(
                safe_call(self.actor_with_delta[(rollout_idx, is_positive)], state)
            )
        else:
            avail_actor_types = '", "'.join(self.get_actor_types())
            raise ValueError(
                f"Invalid parameter type: {actor_type}, "
                f'available options are: "original", "{avail_actor_types}"'
            )

    def store_reward(self, reward: float, actor_type: str, *_, **__):
        """
        Store rollout reward (usually value of the whole rollout episode) for
        each actor type.

        Args:
            reward: Rollout reward.
            actor_type: Actor type.
        """
        if actor_type.startswith("positive_") or actor_type.startswith("negative_"):
            rollout_idx = int(actor_type.split("_")[1])
            is_positive = actor_type[0] == "p"
            self.reward[rollout_idx][is_positive].append(reward)
        else:
            avail_actor_types = '", "'.join(self.get_actor_types())
            raise ValueError(
                f"Invalid parameter type: {actor_type}, "
                f'available options are: "original", "{avail_actor_types}"'
            )

    def update(self):
        """
        Update actor network using rollouts.

        Note:
            All processes in the ARS group must enter this function.
        """
        is_manager = (
            self.ars_group.get_group_members()[0] == self.ars_group.get_cur_name()
        )
        # calculate average reward of collected episodes
        pos_reward, neg_reward, delta_idx = self._get_reward_and_delta()

        # collect result in manager process
        self.ars_group.pair(
            f"ars/rollout_result/{self.ars_group.get_cur_name()}",
            [pos_reward, neg_reward, delta_idx],
        )
        if self.normalize_state:
            self.ars_group.pair(
                f"ars/filter/{self.ars_group.get_cur_name()}", self.filter
            )
        self.ars_group.barrier()

        if is_manager:
            delta_idxs = []  # type: List[Dict[str, int]]
            pos_rewards = []  # type: List[int]
            neg_rewards = []  # type: List[int]
            for m in self.ars_group.get_group_members():
                pos_reward, neg_reward, delta_idx = self.ars_group.get_paired(
                    "ars/rollout_result/" + m
                ).to_here()

                delta_idxs += delta_idx
                pos_rewards += pos_reward
                neg_rewards += neg_reward

            # shape: [2, rollout_num]
            rollout_rewards = np.array([pos_rewards, neg_rewards])
            max_rewards = np.max(rollout_rewards, axis=0)

            # select top performing directions if
            # used_rollout_num < rollout_num
            idx = np.arange(max_rewards.size)[
                max_rewards
                >= np.percentile(
                    max_rewards, 100 * (1 - (self.used_rollout_num / self.rollout_num))
                )
            ]
            delta_idxs = [delta_idxs[i] for i in idx]
            rollout_rewards = rollout_rewards[:, idx]

            # normalize rewards by their standard deviation
            var = np.std(rollout_rewards)
            if not np.isclose(var, 0.0):
                rollout_rewards /= var
            self.actor.zero_grad()
            # aggregate rollouts to form gradient
            # use neg_rollout_rewards - pos_rollout_rewards
            # because -alpha * gradient is added to
            # parameters in SGD
            self._cal_gradient(rollout_rewards[1] - rollout_rewards[0], delta_idxs)
            # nn.utils.clip_grad_norm_(
            #     self.actor.parameters(), self.grad_max
            # )
            # apply gradients
            self.actor_optim.step()

            # collect state statistics
            if self.normalize_state:
                filters = []  # type: List[Dict[str, MeanStdFilter]]
                for m in self.ars_group.get_group_members():
                    filters.append(
                        self.ars_group.get_paired("ars/filter/" + m).to_here()
                    )
                for k, f in self.filter.items():
                    for ff in filters:
                        self.filter[k].collect(ff[k])
                    self.filter[k].apply_stats()
                    self.filter[k].clear_local()

        self.ars_group.barrier()
        self.ars_group.unpair(f"ars/rollout_result/{self.ars_group.get_cur_name()}")
        if self.normalize_state:
            self.ars_group.unpair(f"ars/filter/{self.ars_group.get_cur_name()}")
        self.ars_group.barrier()

        # synchronize filter states across all workers (and the manager)
        if self.normalize_state:
            self._sync_filter()

        # synchronize parameters across all workers (and the manager)
        self._sync_actor()
        # generate new actor parameters with positive and negative delta
        self._generate_parameter()

        # reset reward dict
        self._reset_reward_dict()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()

    def _get_reward_and_delta(self):
        r_range = [i + self.local_rollout_min for i in range(self.local_rollout_num)]
        pos_reward = []
        neg_reward = []
        delta_idx = []
        for i in r_range:
            assert self.reward[i][0] and self.reward[i][1], (
                "You must store rewards for parameters with positive "
                "noise delta and negative noise delta!"
            )
            pos_reward.append(np.mean(self.reward[i][1]))
            neg_reward.append(np.mean(self.reward[i][0]))
            delta_idx.append(self.delta_idx[i])
        return pos_reward, neg_reward, delta_idx

    def _cal_gradient(self, reward_diff: np.array, delta_idxs: List[Dict[str, int]]):
        sampler = SharedNoiseSampler(self.noise_array, 0)
        for name, param in self.actor.named_parameters():
            deltas = [
                sampler.get(delta_idx[name], param.nelement()).reshape(param.shape)
                * r_diff
                for r_diff, delta_idx in zip(reward_diff, delta_idxs)
            ]
            delta = t.mean(
                t.stack(deltas).to(dtype=param.dtype, device=param.device), dim=0
            )

            with t.no_grad():
                param.grad = delta

    def _sync_filter(self):
        is_manager = (
            self.ars_group.get_group_members()[0] == self.ars_group.get_cur_name()
        )
        if is_manager:
            self.ars_group.pair("ars/filter_m", self.filter)
        self.ars_group.barrier()
        if not is_manager:
            manager_filter = self.ars_group.get_paired("ars/filter_m").to_here()
            for k, f in self.filter.items():
                f.sync(manager_filter[k])
                f.apply_stats()
        self.ars_group.barrier()
        if is_manager:
            self.ars_group.unpair("ars/filter_m")
        self.ars_group.barrier()

    def _sync_actor(self):
        is_manager = (
            self.ars_group.get_group_members()[0] == self.ars_group.get_cur_name()
        )
        if is_manager:
            assert self.actor_model_server.push(self.actor), "Push failed"
        self.ars_group.barrier()
        if not is_manager:
            assert self.actor_model_server.pull(self.actor), "Pull failed"
        self.ars_group.barrier()

    def _reset_reward_dict(self):
        self.reward = {}
        for lrn in range(self.local_rollout_num):
            r_idx = lrn + self.local_rollout_min
            self.reward[r_idx] = ([], [])

    def _generate_parameter(self):
        """
        Generate new actor parameters with positive and negative noise deltas.
        """
        self.actor_with_delta = {}
        for lrn in range(self.local_rollout_num):
            r_idx = lrn + self.local_rollout_min
            actor_positive = copy.deepcopy(self.actor)
            actor_negative = copy.deepcopy(self.actor)
            self.delta_idx[r_idx] = {}  # type: Dict[str, int]
            for (name, param), param_p, param_n in zip(
                self.actor.named_parameters(),
                actor_positive.parameters(),
                actor_negative.parameters(),
            ):
                param_size = param.nelement()
                param_sampler = self.noise_sampler[r_idx][name]
                idx, delta = param_sampler.sample(param_size)
                delta = delta.reshape(param.shape).to(
                    dtype=param.dtype, device=param.device
                )
                self.delta_idx[r_idx][name] = idx
                with t.no_grad():
                    param_p.data.copy_(param.data + delta)
                    param_n.data.copy_(param.data - delta)
            self.actor_with_delta[(r_idx, False)] = actor_negative
            self.actor_with_delta[(r_idx, True)] = actor_positive

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "model_server_group_name": "ars_model_server",
            "model_server_members": "all",
            "ars_group_name": "ars",
            "ars_members": "all",
            "models": ["Actor"],
            "model_args": ((),),
            "model_kwargs": ({},),
            "optimizer": "Adam",
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "learning_rate": 0.001,
            "gradient_max": np.inf,
            "noise_std_dev": 0.02,
            "noise_size": 250000000,
            "rollout_num": 32,
            "used_rollout_num": 32,
            "normalize_state": True,
            "noise_seed": 12345,
            "sample_seed": 123,
        }
        config = copy.deepcopy(config)
        config["frame"] = "ARS"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        world = get_world()
        f_config = copy.deepcopy(config["frame_config"])
        ars_group = world.create_rpc_group(
            group_name=f_config["ars_group_name"],
            members=(
                world.get_members()
                if f_config["ars_members"] == "all"
                else f_config["ars_members"]
            ),
        )

        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]

        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        servers = model_server_helper(
            model_num=1,
            group_name=f_config["model_server_group_name"],
            members=f_config["model_server_members"],
        )
        del f_config["optimizer"]
        del f_config["lr_scheduler"]
        frame = cls(
            *models,
            optimizer,
            ars_group,
            servers,
            lr_scheduler=lr_scheduler,
            **f_config,
        )
        return frame



from . import (
    checker,
    conf,
    helper_classes,
    learning_rate,
    logging,
    media,
    prepare,
    save_env,
    tensor_board,
    visualize,
)

__all__ = [
    "checker",
    "conf",
    "helper_classes",
    "learning_rate",
    "logging",
    "media",
    "prepare",
    "save_env",
    "tensor_board",
    "visualize",
]



from threading import Event, Lock
import time

_lock = Lock()


def init_or_register(event, parent_event):
    if hasattr(event, "_magic_parent_events"):
        event._magic_parent_events.append(parent_event)
    else:
        org_set = event.set
        event._magic_parent_events = [parent_event]

        def new_set():
            nonlocal event, org_set
            for pe in event._magic_parent_events:
                with pe._cond:
                    pe._cond.notify_all()
            org_set()

        event.set = new_set


class MultiEvent(Event):
    # can only be used with Event from threading and not from multiprocessing
    def __init__(self, *events):
        global _lock
        super().__init__()

        with _lock:
            self.is_leaf = all([type(e) == Event for e in events])
            for event in events:
                if type(event) == Event:
                    init_or_register(event, self)
                elif isinstance(event, MultiEvent):
                    for le in event.get_leaf_events():
                        init_or_register(le, self)
                else:
                    raise ValueError(
                        f"Type {type(event)} is not a valid event type, "
                        "requires threading.Event or an instance of MultiEvent."
                    )
        self._events = events

    def set(self):
        pass

    def clear(self):
        pass

    def is_set(self):
        return False

    def get_leaf_events(self):
        for event in self._events:
            if type(event) == Event:
                yield event
            else:
                for event in event.get_leaf_events():
                    yield event

    def wait(self, timeout=None):
        begin = time.time()
        while True:
            with self._cond:
                self._cond.wait(timeout)
            if timeout is None or (
                timeout is not None and time.time() - begin >= timeout
            ):
                break
        return self.is_set()


class OrEvent(MultiEvent):
    def is_set(self):
        return any([event.is_set() for event in self._events])


class AndEvent(MultiEvent):
    def is_set(self):
        return all([event.is_set() for event in self._events])



import traceback
import threading


class ThreadException(Exception):
    pass


class Thread(threading.Thread):
    """
    Enhanced thread with exception tracing.
    """

    def __init__(
        self,
        group=None,
        target=None,
        name=None,
        args=(),
        kwargs={},
        cleaner=None,
        *,
        daemon=None,
    ):
        threading.Thread.__init__(
            self,
            group=group,
            target=target,
            name=name,
            args=args,
            kwargs=kwargs,
            daemon=daemon,
        )
        self._cleaner = cleaner
        self._exception_str = ""
        self._has_exception = False

    @property
    def exception(self):
        if not self._has_exception:
            return None
        exc = ThreadException(self._exception_str)
        return exc

    def watch(self):
        if self._has_exception:
            raise self.exception

    @staticmethod
    def format_exceptions(exceptions):
        all_tb = ""
        for exc, i in zip(exceptions, range(len(exceptions))):
            tb = exc.__traceback__
            tb = traceback.format_exception(type(exc), exc, tb)
            tb = "".join(tb)
            all_tb = all_tb + f"\nException {i}:\n{tb}"
        return all_tb

    def run(self):
        exc = []
        try:
            super().run()
        except BaseException as e:
            exc.append(e)
        finally:
            if self._cleaner is not None:
                try:
                    self._cleaner()
                except BaseException as e:
                    exc.append(e)
            if exc:
                self._exception_str = self.format_exceptions(exc)
                self._has_exception = True



import os
import time
import tempfile
import numpy as np
import pytorch_lightning as pl
from typing import Iterable, List, Dict, Union, Any, Tuple, Callable
from torch.utils.data import IterableDataset
from pytorch_lightning.loggers.base import LoggerCollection
from machin.utils.media import create_video, numpy_array_to_pil_image


Scalar = Any


def determine_precision(models):
    dtype = set()
    for model in models:
        for k, v in model.named_parameters():
            dtype.add(v.dtype)
    dtype = list(dtype)
    if len(dtype) > 1:
        raise RuntimeError(
            "Multiple data types of parameters detected "
            f"in models: {dtype}, this is currently not supported "
            "since we need to determine the data type of your "
            "model input from your model parameter data type."
        )
    return dtype[0]


def get_loggers_as_list(module: pl.LightningModule):
    if isinstance(module.logger, LoggerCollection):
        return module.logger._logger_iterable
    else:
        return [module.logger]


def log_image(module, name, image: np.ndarray):
    for logger in get_loggers_as_list(module):
        if hasattr(logger, "log_image") and callable(logger.log_image):
            logger.log_image(name, numpy_array_to_pil_image(image))


def log_video(module, name, video_frames: List[np.ndarray]):
    # create video temp file
    fd, path = tempfile.mkstemp(suffix=".gif")
    os.close(fd)
    try:
        create_video(
            video_frames,
            os.path.dirname(path),
            os.path.basename(os.path.splitext(path)[0]),
            extension=".gif",
        )
    except Exception as e:
        print(e)
        os.remove(path)
        return

    size = os.path.getsize(path)
    while True:
        time.sleep(1)
        new_size = os.path.getsize(path)
        if size != 0 and new_size == size:
            break
        size = new_size

    for logger in get_loggers_as_list(module):
        if hasattr(logger, "log_artifact") and callable(logger.log_artifact):
            logger.log_artifact(path, name + ".gif")
    if os.path.exists(path):
        os.remove(path)


class DatasetResult:
    def __init__(
        self,
        observations: List[Dict[str, Any]] = None,
        logs: List[Dict[str, Union[Scalar, Tuple[Scalar, str]]]] = None,
    ):
        self.observations = observations or []
        self.logs = logs or []

    def add_observation(self, obs: Dict[str, Any]):
        self.observations.append(obs)

    def add_log(self, log: Dict[str, Union[Scalar, Tuple[Any, Callable]]]):
        self.logs.append(log)

    def __len__(self):
        return len(self.observations)


class RLDataset(IterableDataset):
    """
    Base class for all RL Datasets.
    """

    early_stopping_monitor = ""

    def __init__(self, **_kwargs):
        super().__init__()

    def __iter__(self) -> Iterable:
        return self

    def __next__(self):
        raise StopIteration()



from . import envs
from . import config
from . import dataset
from . import launcher
from . import pl_logger
from . import pl_plugin

__all__ = ["env", "config", "dataset", "launcher", "pl_logger", "pl_plugin"]



from typing import Union, List, Dict, Any, Callable
from torch import distributed as dist
from torch.utils.data.dataloader import DataLoader
from .dataset import RLDataset
from .config import init_algorithm_from_config, is_algorithm_distributed
from .pl_plugin import DDPPlugin, DDPSpawnPlugin
from machin.parallel.distributed import get_world, debug_with_process as debug
from machin.frame.algorithms import TorchFramework
from machin.utils.conf import Config
import pytorch_lightning as pl


class Launcher(pl.LightningModule):
    def __init__(
        self,
        config: Union[Dict[str, Any], Config],
        env_dataset_creator: Callable[[TorchFramework, Dict[str, Any]], RLDataset],
    ):
        """
        Launcher module for all algorithm frameworks.

        Args:
            config: All configs, including frame config, train env config, etc.
            env_dataset_creator: A callable which accepts the algorithm frame
                and env config dictionary, and outputs a environment dataset.
        """
        super().__init__()
        self.config = config
        self.env_dataset_creator = env_dataset_creator
        assert not is_algorithm_distributed(config)
        self.frame = init_algorithm_from_config(self.config, model_device=self.device)
        self._frame_pl_inited = False
        self.automatic_optimization = False

        # forward models to the launcher module, so that parameters are handled.
        for name, model in zip(self.frame.get_top_model_names(), self.frame.top_models):
            self.add_module(name, model)

    def on_train_start(self):
        self._init_frame_with_pl()

    def on_test_start(self):
        self._init_frame_with_pl()

    def train_dataloader(self):
        return DataLoader(
            dataset=self.env_dataset_creator(
                self.frame, self.config["train_env_config"]
            ),
            collate_fn=lambda x: x,
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.env_dataset_creator(
                self.frame, self.config["test_env_config"]
            ),
            collate_fn=lambda x: x,
        )

    def training_step(self, batch, _batch_idx):
        for idx, sample in enumerate(batch):
            self.frame.store_episode(sample.observations)
            for _ in range(len(sample.observations)):
                self.frame.update()
            self._log(sample.logs)

    def test_step(self, batch, _batch_idx):
        for sample in batch:
            self._log(sample.logs)

    def configure_optimizers(self):
        return self.frame.optimizers, self.frame.lr_schedulers

    def _init_frame_with_pl(self):
        if not self._frame_pl_inited:
            acc_con = self.trainer.accelerator_connector

            if acc_con.use_ddp:
                raise RuntimeError(
                    f"Current framework: {self.config['frame']} is not a "
                    "distributed framework, you should not use an accelerator."
                )

            # Replace optimizer function with pytorch lightning
            # We can probably support DDP automatically for non-distributed
            # models in this way in the future
            self.frame.set_backward_function(self.manual_backward)
            self.frame.optimizers = (
                self.optimizers()
                if isinstance(self.optimizers(), list)
                else [self.optimizers()]
            )
            self._frame_pl_inited = True

    def _log(self, logs: List[Dict[str, Any]]):
        for log in logs:
            for log_key, log_val in log.items():
                if isinstance(log_val, tuple) and callable(log_val[1]):
                    log_val[1](self, log_key, log_val[0])
                else:
                    self.log(log_key, log_val, prog_bar=True)


class DistributedLauncher(pl.LightningModule):
    def __init__(
        self,
        config: Union[Dict[str, Any], Config],
        env_dataset_creator: Callable[[TorchFramework, Dict[str, Any]], RLDataset],
    ):
        """
        Launcher module for all algorithm frameworks.

        Args:
            config: All configs, including frame config, train env config, etc.
            env_dataset_creator: A callable which accepts the algorithm frame
                and env config dictionary, and outputs a environment dataset.
        """
        super().__init__()
        self.config = config
        self.env_dataset_creator = env_dataset_creator
        assert is_algorithm_distributed(config)
        # Lazy initialization for distributed frameworks, framework must be initialized
        # after initializing world.
        self.frame = None
        self.group = None
        # early stopping monitor
        self._train_esm = None
        self._test_esm = None
        self._frame_pl_inited = False
        self._batch_num = None
        self.automatic_optimization = False

    def on_train_start(self):
        self._init_frame_with_pl()

    def on_test_start(self):
        self._init_frame_with_pl()

    def train_dataloader(self):
        dataset = self.env_dataset_creator(self.frame, self.config["train_env_config"])
        self._train_esm = dataset.early_stopping_monitor
        return DataLoader(
            dataset=dataset, batch_size=self._batch_num, collate_fn=lambda x: x
        )

    def test_dataloader(self):
        dataset = self.env_dataset_creator(self.frame, self.config["test_env_config"])
        self._test_esm = dataset.early_stopping_monitor
        return DataLoader(dataset=dataset, collate_fn=lambda x: x)

    def training_step(self, batch, batch_idx):
        debug(f"Begin training step batch {batch_idx}")
        for idx, sample in enumerate(batch):
            self.frame.store_episode(sample.observations)
            debug("Begin update")
            if self.current_epoch != 0 or batch_idx != 0:
                # Skip the first batch so that buffers can be filled
                for _ in range(len(sample.observations)):
                    self.frame.update()
            debug("End update")
            if idx == 0:
                # only log the first batch result, since different
                # processes may have different number of batches.
                # Otherwise barrier in log will fail
                self._log(sample.logs)
            debug("Logging finished")

        # Make sure they exit together
        self.group.barrier()
        debug(f"End training step batch {batch_idx}")

    def test_step(self, batch, _batch_idx):
        for sample in batch:
            self._log(sample.logs)

    def configure_optimizers(self):
        # Distributed case
        # Sets ``lr_schedulers`` and ``optimizers`` attribute of trainer.accelerator
        # later when framework is initialized by overloaded DDP plugins.
        return None

    def init_frame(self):
        # Called by overloaded pytorch lightning DDP plugins
        if self.frame is None:
            # initialize framework
            self.frame = init_algorithm_from_config(
                self.config, model_device=self.device
            )
            self._batch_num = self.config["batch_num"].get(self.frame.role, 1)
            # forward models to the launcher module, so that parameters are handled.
            for name, model in zip(
                self.frame.get_top_model_names(), self.frame.top_models
            ):
                self.add_module(name, model)

            # create group for custom synchronization with large timeout
            # otherwise the default barrier in pytorch_lightning after training step
            # will throw an exception.
            world = get_world()
            self.group = world.create_collective_group(world.get_ranks(), timeout=86400)

    def _init_frame_with_pl(self):
        if not self._frame_pl_inited:
            acc_con = self.trainer.accelerator_connector

            # For distributed frame, use default settings, don't change optimizers.
            if not acc_con.use_ddp or type(acc_con.training_type_plugin) not in (
                DDPPlugin,
                DDPSpawnPlugin,
            ):
                raise RuntimeError(
                    f"Current framework: {self.config['frame']} is a distributed "
                    "framework, you should initialize the "
                    "trainer with a ddp type accelerator, and "
                    "must import machin.auto package to patch"
                    "the default DDP plugin."
                )

            self._frame_pl_inited = True

    def _log(self, logs: List[Dict[str, Any]]):
        for log in logs:
            for log_key, log_val in log.items():
                if isinstance(log_val, tuple) and callable(log_val[1]):
                    log_val[1](self, log_key, log_val[0])
                else:
                    is_dist_initialized = dist.is_available() and dist.is_initialized()
                    # only reduce the early-stopping monitor
                    # since other keys may not be presented in other processes
                    if log_key in (self._train_esm, self._test_esm):
                        self.log(
                            log_key,
                            log_val,
                            prog_bar=True,
                            sync_dist=is_dist_initialized,
                        )
                    else:
                        self.log(log_key, log_val, prog_bar=True)
